#include <stdio.h>
#include <stdlib.h>

#include "machinedef.h"

#ifdef VAX
/*#include "multinet_root:[multinet.include.sys]types.h"*/
/*#include "multinet_root:[multinet.include.sys]socket.h"*/
/*#include "multinet_root:[multinet.include.netinet]in.h"*/
/*#include "multinet_root:[multinet.include]netdb.h"*/
/*#include "multinet_root:[multinet.include.sys]time.h"*/
#include "[.mnet]types.h"
#include "[.mnet]socket.h"
#include "[.mnet]in.h"
#include "[.mnet]netdb.h"
#include "[.mnet]time.h"
#include <stat.h>
#include <unixio.h>
#include <file.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#endif

#include "xfer.h"

FILE *ttyout;

/******************************************************************************
FUNCTION NAME: initsession
DESCRIPTION: Initialize the client end of the network session.  The host to
   be connected with is specified in 'hostname'.  An optional alternate port
   may also be specified.  This function generates the network address of the
   host, attempts to find a port number using a service name, then attempts
   to connect to the peer server on the remote host.
RETURNS: XFR_NOERR on success, XFR_IS_... on error
INPUT PARAMETERS: s - address of socket stream to be defined
                  hostname - name of host to connect to
                  exename - name of this executable (for error messages)
OUTPUT PARAMETERS: s
CALLED BY: main (x.c)
CALLS: gethostbyname socket bcopy htons getservbyname connect
*/
int initsession(s, hostname, exename)
   int *s;
   char *hostname;
   char *exename;
{
   int result = XFR_NOERR;

   struct sockaddr_in sin;
   struct hostent *hp;
   struct servent *sp;


   hp = gethostbyname(hostname);
   if (hp == NULL)
   {
      fprintf(stderr, "%s: remote machine %s unknown\n", exename, hostname);
      result = XFR_IS_ERR;
   }
   else if ((*s = socket(hp->h_addrtype, SOCK_STREAM, 0)) < 0)
   {
      net_error("socket");
      result = XFR_IS_ERR;
   }
   else
   {
      sin.sin_family = hp->h_addrtype;
      bcopy(hp->h_addr, &sin.sin_addr, hp->h_length);

      sp = (struct servent *) getservbyname(XFR_SERVICE, "tcp");
      if (sp == NULL)
      {
         if (XFR_TRYDEFSRV) 
         { 
            fprintf(stderr, "tcp/%s: unknown service.  Attempting to use service %d\n",
                    XFR_SERVICE, XFR_DEFSRV);
            sin.sin_port = htons(XFR_DEFSRV);
         } 
         else
         {  
            fprintf(stderr, "tcp/%s: unknown service.\n");
            result = XFR_IS_ERR;
         }
      }
      else
         sin.sin_port = htons(sp->s_port);

      if (result == XFR_NOERR)
      {
         if (connect(*s, &sin, sizeof (sin), 0) < 0)
         {
            net_error("connect");
            result = XFR_IS_ERR;
         }
      }
   }

   return((int) result);
} /* initsession */

/******************************************************************************
FUNCTION NAME: verifyuser
DESCRIPTION: Send to the server the name of the user the file will be sent to, 
   as well as the name of the user sending the file.  The server checks out 
   the receiving user on its machine, attemps to setuid to it and to set up
   a transfer directory for that user.  The server sends an acknowledgement back
   indicating the result of the user verification process.
RETURNS: Return value of acknowledgement from server (XFR_NOERR on success, 
   XFR_VF_... on error).
INPUT PARAMETERS: s - socket stream
                  remuserid - remote userid (user on remote machine receiving file)
OUTPUT PARAMETERS: none
CALLED BY: main (x.c)
CALLS: net_putline net_readack (readwrite.c)  sprintf
*/
int verifyuser(s, remuserid)
   int s;
   char *remuserid;
{
   char locuserid[256];
   char buf[PACKETLEN];

#ifdef VAX
   cuserid(locuserid);
#else
   struct passwd *pwent;

   pwent = getpwuid(geteuid());
   strcpy(locuserid, pwent->pw_name);
#endif
   
   
   sprintf(buf, "%s %s %s\n", XFR_CMD_USER, remuserid, locuserid);
   net_putline(s, buf);

   return(net_readack(s));
} /* verifyuser */

/******************************************************************************
FUNCTION NAME: readpack
DESCRIPTION: Read one packet of data from the local file which is being
   transferred.  Place data in specified buffer.
RETURNS: 1 on EOF, 0 otherwise.
INPUT PARAMETERS: f - input file descriptor
                  packet - buffer for the data
                  packetlen - number of bytes actually transferred.
OUTPUT PARAMETERS: packetlen
CALLED BY: transferfile (x.c)
CALLS: read
*/
int readpack(f, packet, packetlen)
   int f;
   char packet[];
   int *packetlen;
{
   int i = 0;
   int ateof = 0;

   *packetlen = read(f, packet, PACKETLEN);

   return(*packetlen == 0);

} /* readpack */

/******************************************************************************
FUNCTION NAME: openfile
DESCRIPTION: Opens a local file for read-only access, and gathers some
   statistics on the file (protection mode, number of disk blocks, file size
   in bytes).  On Sun, the number of blocks is determined from the stat
   function directly.  On VMS, the number of blocks (Sun blocks) is estimated
   from the file size.
RETURNS: XFR_NOERR on success, XFR_OF_... on error
INPUT PARAMETERS: f - address of input file descriptor
                  filename - name of file to be opened
                  filemod - address of protection mode of file
                  fileblk - address of size of file in Sun disk blocks
                  filesiz - address of size of file in bytes
OUTPUT PARAMETERS: f, filemod, fileblk, filesiz
CALLED BY: main (x.c)
CALLS: open stat
*/
int openfile(f, filename, filemod, fileblk, filesiz)
   int *f;
   char *filename;
   long unsigned *filemod;
   long unsigned *fileblk;
   long unsigned *filesiz;
{
   int filelen;
   struct stat fileinfo;

   /* Open file to be transferred. */
   if ((*f = open(filename, O_RDONLY, 0)) >= 0)
   {
      stat(filename, &fileinfo);
#ifdef VAX
      *fileblk = fileinfo.st_size / XFR_DISKBLKSIZ;
#else
      *fileblk = fileinfo.st_blocks / 2;
#endif
      *filesiz = fileinfo.st_size;
      *filemod = fileinfo.st_mode & 0x3ff;
      return(XFR_NOERR);
   }
   else
      return(XFR_OF_COF);
} /* openfile */

/******************************************************************************
FUNCTION NAME: initfile
DESCRIPTION: Initialize the file transfer process with the network server.
   This function sends the 'FILE' command with the name of the file,
   the protection mode, and the file size in blocks and bytes to the server.
   The server first checks to see if its disk has enough room for the file,
   then attempts to create the file.  An acknowledgement is read from the server 
   and the result is returned.
RETURNS: Ackowledgement from server - XFR_NOERR on success, XFR_IF_... on error
INPUT PARAMETERS: s - socket stream
                  filename - name of file to be transferred
                  filemod - protection mode of file
                  fileblk - number of Sun disk blocks the file takes
                  filesiz - size of file in bytes
OUTPUT PARAMETERS: none
CALLED BY: main (x.c)
CALLS: net_putline net_readack (readwrite.c) sprintf
*/
int initfile(s, filename, filemod, fileblk, filesiz)
   int s;
   char *filename;
   long unsigned filemod, fileblk, filesiz;
{
   unsigned char buf[PACKETLEN];
   int indx;
   char newfilename[1024];
   char delim;

#ifdef VAX
   delim = ']';
#else
   delim = '/';
#endif

   indx = strlen(filename) - 1;
   while ((indx >= 0) && (*(filename + indx) != delim))
      indx--;
   strcpy(newfilename, filename + indx + 1);

   sprintf(buf, "%s %s %lu %lu %lu\n", XFR_CMD_FILE, newfilename, filemod, 
           fileblk, filesiz);
   net_putline(s, buf);

   return(net_readack(s));
} /* initfile */

/******************************************************************************
FUNCTION NAME: transferfile
DESCRIPTION: Transmit the entire contents of one file to the server.  The
   file has already been opened for reading on the client and for writing
   on the server.  The server is waiting to receive the data.
   The file is read one packet at a time and sent to the server.  The server
   sends one ackowledgement when the entire file has been received (it bases
   the file length by the number of bytes specified by initfile (FILE command)
   This transfer is prefaced by a 'DATA' command to the server.
RETURNS: Value of error code in ackowledgement from server (XFR_NOERR on success,
   XFR_SD_... on error)
INPUT PARAMETERS: s - socket stream
                  f - input file descriptor
                  packetcnt - address of number of packets transferred.
OUTPUT PARAMETERS: none
CALLED BY: main (x.c)
CALLS: readpack (x.c) net_putline net_putdata net_readack (readwrite.c)
*/
int transferfile(s, f, packetcnt, checksum)
   int s;
   int f;
   unsigned long *packetcnt;
   short unsigned int *checksum;
{
   int error = XFR_NOERR;
   int done = 0;
   char packet[PACKETLEN + 1];
   int packetlen;

   *checksum = 0;

   sprintf(packet, "%s\n", XFR_CMD_DATA);
   net_putline(s, packet);

   *packetcnt = 0;

   /* Transfer the file */
   while (!error && !done)
   {
      done = readpack(f, packet, &packetlen);
      if (!done)
      {
         cksum(packet, packetlen, checksum);
         if (net_putdata(s, packet, packetlen) < 0)
            error = XFR_SD_CPP;
         (*packetcnt)++;
      }
      else
         net_putdata(s, (char *)0, XFR_DATAFLAG);
   }

 
   if (error != XFR_NOERR)
      return(error);
   else
   {
      error = net_readack(s);

      /* Allow the session to continue with a file size discrepency warning */
      if (error == XFR_SD_FSZ)
         return(XFR_NOERR);
      else
         return(error);
   }
} /* transferfile */

int verifychecksum(s, checksum)
   int s;
   short unsigned int checksum;
{
   unsigned char buf[PACKETLEN];

   sprintf(buf, "%s %u\n", XFR_CMD_CKSM, checksum);
   net_putline(s, buf);

   return(net_readack(s));
} /* verifychecksum */

/******************************************************************************
FUNCTION NAME: closesession
DESCRIPTION: Closes session with server by sending it a 'QUIT' command.  An
   ackowledgement is received from the server.  It could fail if the server
   cannot send mail to user receiving the file.
RETURNS: Value of error code in ackowledgement from server (XFR_NOERR on success,
   XFR_CS_... on error)
INPUT PARAMETERS: s - socket stream
OUTPUT PARAMETERS: none
CALLED BY:  main (x.c)
CALLS: net_putline net_readack (readwrite.c)
*/
int closesession(s)
   int s;
{
   char buf[PACKETLEN];

   sprintf(buf, "%s\n", XFR_CMD_QUIT);
   net_putline(s, buf);
   return(net_readack(s));
} /* closesession */

#ifdef VAX
void disablerece()
   { fprintf(stderr, "This function not implemented on VAX/VMS.\n"); }

void enablerece()
   { fprintf(stderr, "This function not implemented on VAX/VMS.\n"); }
#else

void getactstat(homedir, uid)
   char *homedir;
   int *uid;
{
}

void disablerece()
{
   struct passwd *pwent;
   int uid;
   struct stat trnstat;
   char trndir[1024];

   pwent = getpwuid(uid = geteuid());
   sprintf(trndir, "%s/%s", pwent->pw_dir, XFR_TRNDIR);

   if (!chdir(pwent->pw_dir))
   {
      if (!chdir(XFR_TRNDIR))
      {
         if (!stat(".", &trnstat))
         {
            if (trnstat.st_uid == uid)
            {
               if (trnstat.st_mode & XFR_TRNWRACS)
               {
                  if (!chmod(".", trnstat.st_mode & XFR_TRNWRMSK))
                  {
                     printf("Transfers disabled for %s\n", pwent->pw_name);
                  }
                  else
                     fprintf(stderr, "Error - could not disable transfers for %s\n",
                             pwent->pw_name);
               }
               else
                  fprintf(stderr, "Transfers already disabled for %s\n", 
                          pwent->pw_name);
            }
            else
               fprintf(stderr, "Error - %s is not owner of transfer directory %s\n",
                       pwent->pw_name, trndir);
         }
         else
            fprintf(stderr, "Error - could not access transfer directory %s\n",
                    trndir);
      }
      else
         if (stat(XFR_TRNDIR, &trnstat))
            fprintf(stderr, "Transfer directory %s does not exist.  User xfer -e first.\n",
                    trndir);
         else
            fprintf(stderr, "Error - could not access transfer directory %s\n",
                    trndir);
   }
   else
      fprintf(stderr, "Error - could not access home directory %s.\n", pwent->pw_dir);
} /* disablerece */

void enablerece()
{
   struct passwd *pwent;
   int uid;
   struct stat trnstat;
   char trndir[1024];

   pwent = getpwuid(uid = geteuid());
   sprintf(trndir, "%s/%s", pwent->pw_dir, XFR_TRNDIR);

   if (!chdir(pwent->pw_dir))
   {
      if (!chdir(XFR_TRNDIR))
      {
         if (!stat(".", &trnstat))
         {
            if (trnstat.st_uid == uid)
            {
               if (!(trnstat.st_mode & XFR_TRNWRACS))
               {
                  if (!chmod(".", trnstat.st_mode | XFR_TRNWRACS))
                  {
                     printf("Transfers enabled for %s\n", pwent->pw_name);
                  }
                  else
                     fprintf(stderr, "Error - could not enable transfers to %s\n",
                             pwent->pw_name);
               }
               else
                  fprintf(stderr, "Transfers already enabled for %s\n", 
                          pwent->pw_name);
            }
            else
               fprintf(stderr, "Error - %s is not owner of transfer directory %s.\n",
                       pwent->pw_name, trndir);
         }
         else
            fprintf(stderr, "Error - could not access transfer directory %s.\n", 
                    trndir);
      }
      else
      {
         if (stat(XFR_TRNDIR, &trnstat))
         {
            if (!mkdir(XFR_TRNDIR, XFR_TRNMODE))
            {
               printf("Transfer directory %s created with transfers enabled for %s\n", 
                      trndir, pwent->pw_name);
            }
            else
               fprintf(stderr, "Error - could not create transfer direcotry %s\n",
                       trndir);
         }
         else
            fprintf(stderr, "Error - could not access transfer directory %s\n",
                    trndir);
      }
   }
   else
      fprintf(stderr, "Error - could not access home directory %s.\n", pwent->pw_dir);
} /* enablerece */
#endif

/******************************************************************************
FUNCTION NAME: parsedest
DESCRIPTION:  Parse command line argument 1 for userid and host destination.  If
   no host name is specified, then the local host name is used.
RETURNS: nothing
INPUT PARAMETERS: dest - combined userid and host destination, in form userid@host
                  userid - userid extracted from dest
                  hostname - host name extracted from dest
OUTPUT PARAMETERS: userid, hostname
CALLED BY: main (x.c)
CALLS: 
*/
void parsedest(dest, userid, hostname)
   char *dest;
   char *userid;
   char *hostname;
{
   int seperator;

   if ((seperator = strchr(dest, '@')) == 0)
   {
      strcpy(userid, dest);
      gethostname(hostname, 256);
   }
   else
   {
      strncpy(userid, dest, (seperator - (int) dest));
      *(userid + (seperator - (int) dest)) = 0;
      strcpy(hostname, (char *) seperator + 1);
   }
} /* parsedest */


   
/******************************************************************************
FUNCTION NAME: main
DESCRIPTION: Main program for client - initializes a network session, verifies the
   remote userid, then opens, sends, and closes the files to be transferred,
   then ends the session with the server.
RETURNS: 0 on success, or error code of any failure
INPUT PARAMETERS: argc, argv - command line parameters
CALLS: initsession verifyuser openfile transferfile closesession (x.c)  close
*/
main(argc, argv)
   int argc;
   char *argv[];
{
   int s;
   char hostname[256];
   char userid[256];
   char *exename;
   char *filename;
   int i;
   int f;
   int packetcnt;
   long unsigned filemod, fileblk, filesiz;
   short unsigned int checksum;

   int error;

#ifdef VAX
   exename = "xfer";
#else
   exename = argv[0];
#endif
   ttyout = stdout;

   if ((argc != 2) && (argc < 3))
   {
      fprintf(stderr, "usage: %s userid[@hostname] file1 [file2, file3, ...]\n", exename);
      fprintf(stderr, 
              "       %s -d Disable receipt of transfers. (Not implemented on VMS)\n",
              exename);
      fprintf(stderr, 
              "       %s -e Enable receipt of transfers. (Not implemented on VMS)\n",
              exename);
      exit(0);
   }

   for (i = 1; *(argv + i); i++)
   { 
      if (!strcmp("-d", *(argv + i)))
      {
         disablerece();
         exit(0);
      }
      else if (!strcmp("-e", *(argv + i)))
      {
         enablerece();
         exit(0);
      }
   } 

   parsedest(argv[1], userid, hostname);

   error = initsession(&s, hostname, exename);

   if (!error)
      error = verifyuser(s, userid);

   for (i = 2; (filename = *(argv + i)) && !error; i++)
   {
      if (error = openfile(&f, filename, &filemod, &fileblk, &filesiz))
         fprintf(stderr, "%s: Error - could not access file %s\n", exename, filename);

      if (!error)
         error = initfile(s, filename, filemod, fileblk, filesiz);
 
      if (!error)
      {
         if (!(error = transferfile(s, f, &packetcnt, &checksum)))
            error = verifychecksum(s, checksum);
         close(f);
      }
      
 
      if (!error)
         printf("Transmitted file %s - %lu packets, %lu bytes, checksum %u.\n",
                filename, packetcnt, filesiz, checksum);
   }
 
   if (!error || (error == XFR_OF_COF))
      error = closesession(s);

   if (error)
      fprintf(stderr, "Errors occurred.  Disconnecting.\n");

   exit(0);
} /* main */
